<?php 

/**
	Joomla:Sber ID Handler	 
	v 1.1.1
 */
namespace Sberbank\Auth;

use DateTime;

define('SBERBANK_ID_SETIFICATE_DIR',  dirname(dirname(__FILE__)) . '/sert/');

class Handler
{
	private $api_version = 'v5'; // 2 || 3

	private $gate_urls = [];

	private $api = [
		'v2' =>  [
			'BASE_URL' => 'https://id.sber.ru',
			// 'BASE_URL' => 'https://online.sberbank.ru',
			'AUTH' => 'https://online.sberbank.ru/CSAFront/oidc/authorize.do',
			'TOKEN' => 'https://api.sberbank.ru/ru/prod/tokens/v2/oidc',
			'USERINFO' => 'https://api.sberbank.ru/ru/prod/sberbankid/v2.1/userinfo',
			'COMPLETED' => ' https://api.sberbank.ru/prod/sberbankid/api/auth/completed',
		],
		'v3' => [
			'BASE_URL' => 'https://id.sber.ru',
			// 'BASE_URL' => 'https://online.sberbank.ru',
			'AUTH' => 'https://online.sberbank.ru/CSAFront/oidc/authorize.do',
			'TOKEN' => 'https://api.sberbank.ru:8443/prod/tokens/v2/oidc',
			'USERINFO' => 'https://api.sberbank.ru/ru/prod/sberbankid/v2.1/userinfo/v3',
			'COMPLETED' => ' https://api.sberbank.ru/prod/sberbankid/api/auth/completed',
		],
		'v5' => [
			'BASE_URL' => 'https://id.sber.ru',
			// 'BASE_URL' => 'https://online.sberbank.ru',
			'AUTH' => 'https://online.sberbank.ru/CSAFront/oidc/authorize.do',
			'TOKEN' => 'https://oauth.sber.ru/ru/prod/tokens/v2/oidc',
			'USERINFO' => 'https://oauth.sber.ru/ru/prod/sberbankid/v2.1/userinfo',
			'COMPLETED' => ' https://oauth.sber.ru/ru/prod/sberbankid/api/auth/completed',
		],
		'TEST' => [
			'BASE_URL' => 'https://id-psi.sber.ru',
			'AUTH' => 'https://id-psi.sber.ru/CSAFront/oidc/authorize.do',
			'TOKEN' => 'https://oauth-psi.sber.ru/ru/prod/tokens/v2/oidc',
			'USERINFO' => 'https://oauth-psi.sber.ru/ru/prod/sberbankid/v2.1/userinfo',
			'COMPLETED' => 'https://oauth-psi.sber.ru/ru/prod/sberbankid/api/auth/completed',
		],
	];

	private $options = [
		'client_id' => false,
		'client_secret' => false,
		'sertificate_file' => false,
		'sertificate_password' => false,
		'sertificate_dir' => false,
		'nonce' => false,
		'state' => false,
		'code' => false,
		'access_token' => false,
		'scope' => 'openid name email',
		'redirect_url' => false,
		'pkce' => false,
		'web_to_app' => false,
		'code_verifier' => false,
		'code_challenge' => false,
		'debug_mode' => false,
		'log_dir' => false,
		'oauth_generated_params' => false,
		'api_version' => 'v5',
		'api_mode' => 'TEST' // TEST | PROM
	];
	private $sertificate_status = [
		'UPLOADED' => false,
		'DECODED' => false,
		'NEW_UPLOAD' => 0,
	];
	private $test_sertificate_status = [
		'UPLOADED' => false,
		'DECODED' => false,
		'NEW_UPLOAD' => 0,
	];
	private $debugData = array();
	private $userinfo = array();
	private $oauth_url = false;
	private $errors = false;
	

	public function __construct($api_version = 'v5',$api_mode = 'TEST') {


		if($api_mode == 'TEST') {
			$this->options['api_mode'] = 'TEST';
			$this->gate_urls = $this->api['TEST'];
		} else {
			$this->options['api_mode'] = 'PROM';
			$this->gate_urls = $this->api[$this->options['api_version']];	
		}

		$this->options['sertificate_dir'] = SBERBANK_ID_SETIFICATE_DIR;
		$this->options['nonce'] = 'n-' . $this->generateRandomString(3) . '_' . $this->generateRandomString(10);
		$this->options['state'] = $this->generateRandomString();

		$protocol = isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == "on" || ( $_SERVER['HTTPS'] == '1') ? 'https://' : 'http://';
		$domain_name = $_SERVER['HTTP_HOST'];
		$this->options['redirect_url'] =  $protocol . $domain_name .'/?option=com_sberid';

		$this->sertificate_status['UPLOADED'] = $this->sertificate_check_exist();
		$this->sertificate_status['DECODED'] = $this->sertificate_check_decoded();
	}

	// public function setGateurl($data) {
	// 	$this->$gate_urls = $data;
	// }

	
	public function setClientId($value) {
		$this->options['client_id'] = $value;
	}


	public function setClientSecret($value) {
		$this->options['client_secret'] = $value;
	}


	public function setSertificatePassword($value) {
		$this->options['sertificate_password'] = $value;
	}


	public function setSertificateFile($value) {
		$this->options['sertificate_file'] = $value;
	}


	public function setRedirectUrl($value) {
		 $this->options['redirect_url'] = $value;
	}


	public function setSertificateDir($value) {
		$this->options['sertificate_dir'] = $value;
	}


	public function setLogDir($value) {
		$this->options['log_dir'] = $value;
	}


	public function getSertificateDir() {
		if(!$this->options['sertificate_dir']) {
			return false;
		}
		if(!file_exists($this->options['sertificate_dir'])) {
			mkdir($this->options['sertificate_dir'], 0700);
		}
		if(!file_exists($this->options['sertificate_dir'] . 'test/')) {
			mkdir($this->options['sertificate_dir'] . 'test/', 0700);
		}
		$test_dir = $this->options['api_mode'] === 'TEST' ? 'test/' : '';
		return $this->options['sertificate_dir'] . $test_dir;
	}
	public function getBaseUrl() {
		return $this->gate_urls['BASE_URL'];
	}

	public function setProperty($prop,$value) {
		if($prop == 'code') {
			$this->options['code'] = $value;
			return true;
		} else if($prop == 'access_token') {
			$this->options['access_token'] = $value;
			return true;
		} else if($prop == 'pkce') {
			$this->options['pkce'] = $value;
			return true;
		} else if($prop == 'web_to_app') {
			$this->options['web_to_app'] = $value;
			return true;
		} else if($prop == 'code_verifier') {
			$this->options['code_verifier'] = $value;
			return true;
		} else if($prop == 'scope') {
			$this->options['scope'] = str_replace("+", " ", $value);
			return true;
		} else if($prop == 'api_mode') {
			$this->options['api_mode'] = $value;
			return true;
		}
		return false;
	}


	public function getProperty($prop) {
		if($prop == 'code') {
			return $this->options['code'];
		} else if($prop == 'access_token') {
			return $this->options['access_token'];
		} else if($prop == 'pkce') {
			return $this->options['pkce'];
		} else if($prop == 'web_to_app') {
			return $this->options['web_to_app'];
		} else if($prop == 'code_verifier') {
			return $this->options['code_verifier'];
		} else if($prop == 'redirect_url') {
			return $this->options['redirect_url'];
		}
		return false;
	}


	public function getProperties() {
		return $this->options;
	}


	public function getUserinfo() {
		return $this->userinfo;
	}


	public function debugOn() {
		$this->options['debug_mode'] = true;
	}


	public function generate_oauth_url() {
		if(substr($this->options['scope'], -1) == '+') {
			$this->options['scope'] = substr($this->options['scope'], 0, -1);
		}
		$auth_url = $this->gate_urls['AUTH'] .
		'?response_type=code' .
		'&client_type=PRIVATE' .
		'&scope=' . $this->options['scope'] .
		'&client_id=' . urlencode($this->options['client_id']) .
		'&redirect_uri=' . urlencode($this->options['redirect_url']) .
		($this->options['nonce'] <> '' ? '&nonce=' . urlencode($this->options['nonce']) : '') .
		($this->options['state'] <> '' ? '&state=' . $this->options['state'] : '');

		$this->options['oauth_generated_params'] = [
			'url' => $this->gate_urls['AUTH'],
			'scope' => $this->options['scope'],
			'client_id' => urlencode($this->options['client_id']),
			'redirect_uri' =>  urlencode($this->options['redirect_url']),
			'nonce' => urlencode($this->options['nonce']),
			'state' => urlencode($this->options['state']),
			'redirect_uri_base' => $this->options['redirect_url'],
			'base_url' => $this->gate_urls['BASE_URL'],
		];
		
		if($this->options['pkce']) {
			$code_verifier = null;
			$code_challenge = null;
			$random = bin2hex(openssl_random_pseudo_bytes(32));
			if($code_verifier === null) {
				$code_verifier = $this->base64url_encode(pack('H*', $random));
			}
			if($code_challenge === null) {
				$code_challenge = $this->base64url_encode(pack('H*', hash('sha256', $code_verifier)));
			}
			$this->options['code_verifier'] = $code_verifier;

			$auth_url .= '&code_challenge=' . $code_challenge;
			$auth_url .= '&code_challenge_method=S256';
		}

		array_push($this->debugData, array(
			'METHOD' => 'GENERATE_AUTH_LINK',
			'AUTH_URL' => $auth_url, 
		));
		return $auth_url;
  
	}


	public function get_oauth_params() {
		return $this->options['oauth_generated_params'];
	}


	public function reqGetAccessToken() {

		$this->checkFileSertificate();
		if($this->errors) {
			return false;
		}
		if(substr($this->options['scope'], -1) == '+') {
			$this->options['scope'] = substr($this->options['scope'], 0, -1);
		}
		$headers = array(
			'accept:' . 'application/json',
	    	'content-type:' . 'application/x-www-form-urlencoded',
	    	'rquid:' . $this->generateRandomString(),
	    	'x-ibm-client-id: ' . $this->options['client_id'],
		);
		$data = array(
			'grant_type' => 'authorization_code',
			'scope' => $this->options['scope'],
			'client_id' => $this->options['client_id'],
			'client_secret' => $this->options['client_secret'],
			'code' => $this->options['code'],
			'redirect_uri' => $this->options['redirect_url'],
		);
		if($this->options['code_verifier']) {
			$data['code_verifier'] = $this->options['code_verifier'];
		}
		
		$response = $this->gateway( $headers, $data, $this->gate_urls['TOKEN'] );
		
		array_push($this->debugData, array(
			'METHOD' => 'GET_ACCESS_TOKEN',
			'REQUEST_URL'=> $this->gate_urls['TOKEN'], 
			'HEADERS'=> $headers,
			'REQUEST_DATA' => $data, 
			'RESPONSE' => $response
		));

		if(isset($response['access_token'])) {
			$this->setProperty( 'access_token', $response['access_token'] );
		} else {
			$this->errors = true;
			return false;
		}

		return true;
	}

	public function reqGetUserInfo() {

		$this->checkFileSertificate();
		if($this->errors) {
			return false;
		}
		$headers = array(
		    'accept:' . 'application/json',
		    'x-ibm-client-id: ' . $this->options['client_id'],
		    'x-introspect-rquid:' . $this->generateRandomString(),
		    'Authorization:Bearer ' . $this->options['access_token'],
		);
		
		$data = array();

		$response = $this->gateway( $headers, $data, $this->gate_urls['USERINFO'] );

		array_push($this->debugData, array(
			'METHOD' => 'GET_USER_INFO',
			'REQUEST_URL'=> $this->gate_urls['USERINFO'], 
			'HEADERS'=> $headers,
			'REQUEST_DATA' => $data, 
			'RESPONSE' => $response
		));
		if(isset($response['httpCode'])) {
			$this->errors = true;	
			return false;
		}
		$this->userinfo = $response;
		return true;
		
	}

	public function reqComplete() {

		
		if($this->errors) {
			return false;
		}
		if($this->options['api_mode'] != 'PROM') {
			return false;	
		}
		
		$headers = array(
	    	// 'content-type:' . 'application/x-www-form-urlencoded',
	    	'rquid:' . $this->generateRandomString(),
	    	'x-ibm-client-id: ' . $this->options['client_id'],
	    	'Authorization:Bearer ' . $this->options['access_token'],
		);

		$data = array();
		
		$response = $this->gateway( $headers, $data, $this->gate_urls['COMPLETED'] );
		
		array_push($this->debugData, array(
			'METHOD' => 'COMPLETED',
			'REQUEST_URL'=> $this->gate_urls['COMPLETED'], 
			'HEADERS'=> $headers,
			'REQUEST_DATA' => $data, 
			'RESPONSE' => $response
		));
		$this->logger();

		return true;
	}

	public function generateRandomString($length = 32) {
	    $characters = 'ABCDEFabcdef0123456789';
	    $charactersLength = strlen($characters);
	    $randomString = '';
	    for ($i = 0; $i < $length; $i++) {
	        $randomString .= $characters[rand(0, $charactersLength - 1)];
	    }
	    return $randomString;
	}

	public function sertificate_save($uploaded_file) {
		if(!$this->getSertificateDir()) {
			return false;
		}
		$file_parts = pathinfo($uploaded_file['name']);
		if($file_parts['extension'] != 'p12') {
			$this->sertificate_status['NEW_UPLOAD'] = 1;
			return false;
		}
		move_uploaded_file($uploaded_file['tmp_name'], $this->getSertificateDir() . 'sertificate.p12');

		$this->sertificate_status['NEW_UPLOAD'] = 2;
		$this->sertificate_status['UPLOADED'] = 1;
		return true;
	}
	public function sertificate_crop($password) {
		if(!$this->getSertificateDir()) {
			return false;
		}
		
		$sert_file_path = $this->getSertificateDir() . 'sertificate.p12';

		if(!file_exists($sert_file_path)) {
			return false;
		}

		$cert_store = file_get_contents($sert_file_path);
		$sert_read_status = openssl_pkcs12_read($cert_store, $cert_info, $password);
		if($sert_read_status) {
			$fp = fopen($this->getSertificateDir() . "client.crt", "w");
			fwrite($fp, $cert_info['cert'].$cert_info['extracerts'][0].$cert_info['extracerts'][1]);
			fclose($fp);

			$fp = fopen($this->getSertificateDir() . "client.key", "w");
			fwrite($fp, $cert_info['pkey']);
			fclose($fp);
			$this->sertificate_status['DECODED'] = $this->sertificate_check_decoded();
			return true;

		}

		return false;
	}
	public function sertificate_check_exist() {
		$file = $this->getSertificateDir() . 'sertificate.p12';
		if(file_exists($file) && filesize($file) > 1) {
			return 1;
		}
		return 0;
	}

	public function sertificate_check_decoded() {
		$file = $this->getSertificateDir() . 'client.crt';
		if(file_exists($file) && filesize($file) > 1) {
			return 1;
		}
		return 0;
	}

	public function sertificate_load_status() {
		return $this->sertificate_status;
	}

	public function base64url_encode($plainText)
	{
	    $base64 = base64_encode($plainText);
	    $base64 = trim($base64, "=");
	    $base64url = strtr($base64, '+/', '-_');
	    return ($base64url);
	}


	public function showDebug($method = false) {
	}

	public function authSuccess() {
		if(!$this->errors) {
			return true;
		}
		return false;
	}

	public function showError($method = false) {

		if($method) {
			return array('errorMessage' => 'ОШИБКА!', 'debugData' => $this->debugData);
		} else {
			echo '<div style="font-family:arial;margin:20px 10px;">';
            echo '<br><b>Ошибка авторизации, обратитесь к администратору!</b><br>' . '' . '<br>' ;
            echo '<a href="/">Вернуться на главную</a>';
            echo "</div>";
           	if($this->options['debug_mode']) {
				echo '<div style="margin:20px 10px;padding:20px;border:#ccc 1px solid; font-family:arial;"><span style="color:red">Включен режим отладки:</span><br>';
				echo "<pre>";
					print_r($this->debugData);
				echo "</pre>";
				
			}
			die;
		}

    }
    public function authSetError($data = false) {
    	if($data) {
    		array_push($this->debugData,$data);
    	}
    	$this->errors = true;
    }

	private function gateway($headers = array(),$data = array(), $gate_url = '') {

		$curl = curl_init();

		curl_setopt_array($curl, array(
		    CURLOPT_URL => $gate_url,
		    CURLOPT_RETURNTRANSFER => true,
		    // CURLOPT_HEADER => true,
		    CURLOPT_HTTPHEADER => $headers,
		    CURLOPT_SSLVERSION => CURL_SSLVERSION_TLSv1_2,
			CURLOPT_RETURNTRANSFER => true,
			CURLOPT_SSLCERT => $this->getSertificateDir() . 'client.crt',
			CURLOPT_SSLKEY => $this->getSertificateDir() . 'client.key',
			CURLOPT_CAINFO => $this->getSertificateDir() . 'client.crt',
			// CURLOPT_SSL_VERIFYHOST => false,
			CURLOPT_SSL_VERIFYPEER => false,
			CURLOPT_SSLKEYPASSWD => $this->options['sertificate_password'],
		));

		if(count($data) > 0) {
			curl_setopt($curl, CURLOPT_POSTFIELDS, http_build_query($data));
		}


		$response = curl_exec($curl);

		$response = json_decode( $response, true );
		// print_r(curl_error($curl));
		// print_r($response);
		curl_close($curl);
		return $response;
	}


	// private function gateway2($headers = array(),$data = array(), $gate_url = '') {

	// 	$curl = curl_init();

	// 	curl_setopt_array($curl, array(
	// 	    CURLOPT_URL => $gate_url,
	// 	    CURLOPT_RETURNTRANSFER => true,
	// 	    CURLOPT_HEADER => true,
	// 	    CURLOPT_HTTPHEADER => $headers,
	// 	    CURLOPT_SSLVERSION => CURL_SSLVERSION_TLSv1_2,
	// 		CURLOPT_RETURNTRANSFER => true,
	// 		CURLOPT_SSLCERT => $this->getSertificateDir() . 'client.crt',
	// 		CURLOPT_SSLKEY => $this->getSertificateDir() . 'client.key',
	// 		CURLOPT_CAINFO => $this->getSertificateDir() . 'client.crt',
	// 		// CURLOPT_SSL_VERIFYHOST => false,
	// 		CURLOPT_SSL_VERIFYPEER => false,
	// 		CURLOPT_SSLKEYPASSWD => $this->options['sertificate_password'],

	// 	));
	// 	if(count($data) > 0) {
	// 		curl_setopt($curl, CURLOPT_POSTFIELDS, http_build_query($data));
	// 	}
		
	// 	$response = curl_exec($curl);
	// 	print_r($gate_url);
	// 	echo "<br>";
	// 	$header_size = curl_getinfo($curl, CURLINFO_HEADER_SIZE);
	// 	$headers = substr($response, 0, $header_size);
	// 	echo "Headers:\n" . $headers . "\n\n";
	// 	$httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
	// 	echo "HTTP Status Code: " . $httpCode . "\n\n";
	// 	$response = json_decode( $response, true );
	// 	// print_r(curl_error($curl));
	// 	// print_r($response);
	// 	curl_close($curl);
	// 	return $response;
	// }

    private function checkFileSertificate() {
    	if($this->errors) {
    		return false;
    	}
    	$fileSert = $this->getSertificateDir() . 'client.crt';
    	
    	if(!$this->getSertificateDir()) {
    		$this->authSetError(array(
    			'METHOD' => 'SERTIFICATE_DIR_EXIST',
    			'MESSAGE' => 'Sertificate directory not specified!'
    		));
    		return false;
    	}
    	if (!file_exists($fileSert)) {
    		$this->authSetError(array(
    			'METHOD' => 'SERTIFICATE_FILE_EXIST',
    			'MESSAGE' => 'Sertificate file not found!'
    		));
    		return false;
        }
        if(filesize($fileSert) <= 1) {
        	$this->authSetError(array(
    			'METHOD' => 'SERTIFICATE_VERIFY',
    			'MESSAGE' => 'Sertificate password not alowed!'
    		));
    		return false;
        }
        return true;
    			        
    }

	public function logger() {

		if($this->options['log_dir']) {
			if(!file_exists($this->options['log_dir'])) {
				mkdir($this->options['log_dir'], 0700);
			}
	        $objDateTime = new DateTime();
	        $file = $this->options['log_dir'] . 'sber_id.log';   
	        $logContent = '';

	        if(file_exists($file)) {
	            $logSize = filesize($file) / 1000;
	            if($logSize < 5000) {
	                $logContent = file_get_contents($file);
	            }
	        }

	        $logContent .= '----------------------------' . "\n";
	        $logContent .= "DATE: " . $objDateTime->format("Y-m-d H:i:s") . "\n";
	        $logContent .= 'API_VERSION: ' . $this->api_version . "\n";
	        $logContent .=  print_r($this->debugData,true);
	        $logContent .= "\n\n";

	        file_put_contents($file, $logContent);		
		}

    }

    public function checkServer() {
    	echo "<pre>";
		$server_info = array();
		$server_info[] = array("PHP version:", phpversion() );
	    if (function_exists('curl_version')) {
	        $curl = curl_version();
	        $server_info[] = array("cURL version:", $curl["version"] );
	        $ch = curl_init('https://www.howsmyssl.com/a/check');
	        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
	        $data = curl_exec($ch);
	        curl_close($ch);
	        $json = json_decode($data);
	        $server_info[] = array("TLS version: ", $json->tls_version );
	    } else {
	    	$server_info[] = array("PHP version:", phpversion() );
	    	$server_info[] = array("cURL", 'Not installed!!!' );
	    }
	    $server_info[] = array("OpenSSL version text: ", OPENSSL_VERSION_TEXT );
	    $server_info[] = array("OpenSSL version number: ", OPENSSL_VERSION_NUMBER );
	    print_r($server_info);
	    die;
    }
    public function test() {
    	echo "sber id is running";
    }
}

?>